###############
# App windows #
###############


proc :cursor {subcmd cursor args} {
  switch -- $subcmd {
    set {
      ::TkEsweepXYPlot::setCursorByValue IMD $cursor [lindex $args 0]
    }
    trace {
      switch -- [set arg [lindex $args 0]] {
        get {
          return -code ok #$::TkEsweepXYPlot::plots(IMD,Config,Cursors,$cursor,Trace)
        }
        default {
          set id [isID $arg]
          set ::TkEsweepXYPlot::plots(IMD,Config,Cursors,$cursor,Trace) $id
          set ::TkEsweepXYPlot::plots(IMD,Config,Cursors,$cursor,Color) [getProp $id Color]
          set ::TkEsweepXYPlot::plots(IMD,Config,Cursors,$cursor,Prefix) "$arg: "
          ::TkEsweepXYPlot::plot IMD
        }
      }
    }
    default {
      return -code error "Unknown subcommand $subcmd"
    }
  }
}

# ---------- bindings
proc bindings {win} {
  bind $win <Escape> [list ::audio::cancel]

  # program control commands
  bind $win <m> [list apply {{} {
      set ID [:measure]
      :eval $ID
    }}]

  # overlay controls
  bind $win <Double-d> [list apply {{} {
      global master
      set IDs [::overlay::select get]
      if {[lsearch $IDs $master] < 0} {
        ::overlay::yank
        foreach ID $IDs {
          ::TkEsweepXYPlot::removeTrace IMD $ID
        }
        ::TkEsweepXYPlot::plot IMD
      } else {
        ::cmdbar::writeToCmdBar {Master curve may not be deleted} -bg red
      }
    }}]
  bind $win <Double-y> [list apply {{} {
      set sel [::overlay::copy]
      if {[llength $sel] == 1} {
        ::cmdbar::writeToCmdBar "Overlay #$sel copied to buffer" -bg green
      } elseif {[llength $sel] > 1} {
        ::cmdbar::writeToCmdBar "Copied [llength $sel] overlays to buffer" -bg green
      } else {
        ::cmdbar::writeToCmdBar {Nothing to copy} -bg green
      }}}]
  bind $win <p> [list apply {{} {
      foreach ID [::overlay::paste] {
        set props [list Name "Copy of [getProp $ID Name]"]
        ::overlay::modify $ID $props
        catch {updateIMD}
      }
    }}]
  # show selected overlays
  bind $win <S> [list apply {{} {
      foreach ID [::overlay::select get] {
        ::overlay::modify $ID [list Show {yes}]
        ::overlay::enable $ID
        if {[::TkEsweepXYPlot::exists IMD $ID]} {
          ::TkEsweepXYPlot::configTrace IMD $ID -state on
          ::TkEsweepXYPlot::plot IMD
        } else {
          catch {updateIMD}
        }
      }
    }}]
  bind $win <s> [bind $win <S>]
  # hide selected overlays
  bind $win <H> [list apply {{} {
      foreach ID [::overlay::select get] {
        ::overlay::modify $ID [list Show {no}]
        ::overlay::disable $ID
        if {[::TkEsweepXYPlot::exists IMD $ID]} {
          ::TkEsweepXYPlot::configTrace IMD $ID -state off
          ::TkEsweepXYPlot::plot IMD
        }
      }
    }}]
  bind $win <h> [bind $win <H>]
   
  # change color
  bind $win <C> [list apply {{win} {
    foreach ID [::overlay::select get] {
      if {[set color [tk_chooseColor -initialcolor [getProp $ID Color] \
                      -parent $win \
                      -title [getProp $ID Name]]] ne {}} {
        ::overlay::modify $ID [list Color $color]
      }
    }
    catch {updateIMD}
   }} $win]
   bind $win <c> [bind $win <C>]

   # rename
   bind $win <R> [list apply {{} {:rename}}]
   bind $win <r> [bind $win <R>]
   bind $win <F2> [bind $win <R>]

   # make selected overlay the master
   bind $win <M> [list apply {{} {
       global master
       set master [lindex [::overlay::select get] 0]
     }}]
   bind $win <Control-m> [bind $win <M>]

   bind $win <period> [list apply {{} {
       ::cmdbar::execCmdBar [::cmdbar::historyGet]
     }
   }]

}


# ---------------- plots
proc createPlots {imd} {
  global master
  ::TkEsweepXYPlot::init $imd IMD
  set ::TkEsweepXYPlot::plots(IMD,Config,Title) {Frequency response and distortion}

  set ::TkEsweepXYPlot::plots(IMD,Config,Font,Size) {-12}

  set ::TkEsweepXYPlot::plots(IMD,Config,Margin,Top) 80
  set ::TkEsweepXYPlot::plots(IMD,Config,Margin,Bottom) 80
  set ::TkEsweepXYPlot::plots(IMD,Config,Margin,Left) 80
  set ::TkEsweepXYPlot::plots(IMD,Config,Margin,Right) 80
  set ::TkEsweepXYPlot::plots(IMD,Config,MouseHoverActive) {no}
  set ::TkEsweepXYPlot::plots(IMD,Config,Legend,Align) center
  set ::TkEsweepXYPlot::plots(IMD,Config,Cursors,1) {on}
  set ::TkEsweepXYPlot::plots(IMD,Config,Cursors,1,Trace) $::master
  set ::TkEsweepXYPlot::plots(IMD,Config,Cursors,2) {off}

  set ::TkEsweepXYPlot::plots(IMD,Config,X,Label) {f [Hz]}
  set ::TkEsweepXYPlot::plots(IMD,Config,X,Precision) 6
  set ::TkEsweepXYPlot::plots(IMD,Config,X,Autoscale) {off}
  set ::TkEsweepXYPlot::plots(IMD,Config,X,Min) 20
  set ::TkEsweepXYPlot::plots(IMD,Config,X,Max) 20e3
  set ::TkEsweepXYPlot::plots(IMD,Config,X,Log) {yes}

  set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Label) {Magnitude [dBSPL]}
  set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Precision) 3
  set ::TkEsweepXYPlot::plots(IMD,Config,Y1,MStep) 5

  set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Autoscale) {on}
  lassign [setup cget Display,Range] ::TkEsweepXYPlot::plots(IMD,Config,Y1,Min) ::TkEsweepXYPlot::plots(IMD,Config,Y1,Max)
  set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Log) {no}

  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,State) {off}
  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,Label) {Phase []}
  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,PolarFormat) {deg}
  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,Precision) 3
  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,MStep) 5
  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,Min) -180
  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,Max) 180
#  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,Autoscale) {on}
#  set ::TkEsweepXYPlot::plots(IMD,Config,Y2,Clamp) [list -180.0 180.0]

  updateIMD

  ::TkEsweepXYPlot::plot IMD
}

proc updateIMD {} {
  set mode [lindex [setup cget Display,Scaling] 0]
  set flags [lrange [setup cget Display,Scaling] 1 end]
  switch -- $mode {
    manual {
      set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Autoscale) {off}
      set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Min) [lindex [setup cget Display,Range] 0]
      set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Max) [lindex [setup cget Display,Range] 1]
    }
    auto -
    default {
      set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Autoscale) {on}
      if {[llength [setup cget Display,Bounds]] == 2} {
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Bounds) [setup cget Display,Bounds]
      } else {
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Bounds) {}
      }
    }
  }
  # save current stats
  set bounds $::TkEsweepXYPlot::plots(IMD,Config,Y1,Bounds)
  set clamp $::TkEsweepXYPlot::plots(IMD,Config,Y1,Clamp)
  set range [list $::TkEsweepXYPlot::plots(IMD,Config,Y1,Max) $::TkEsweepXYPlot::plots(IMD,Config,Y1,Max)]
  foreach flag $flags {
    switch -- $flag {
      base0 {
        # set the lower bound to 0
        if {$::TkEsweepXYPlot::plots(IMD,Config,Y1,Autoscale) eq {on}} {
          lset ::TkEsweepXYPlot::plots(IMD,Config,Y1,Clamp) 0 0
        } else {
          set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Min) 0
        }

        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Log) no
      }
      lin {
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Log) no
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Min) [lindex $range 0]
      }
      log {
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Log) yes
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Min) [lindex $range 0]
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Bounds) $bounds
        set ::TkEsweepXYPlot::plots(IMD,Config,Y1,Clamp) $clamp
      }
    }
  }

  set IDs [::overlay::ids]
  foreach id $IDs {
    if {[getProp $id Show] eq {yes}} {
      showIMD $id noplot
    }
  }

  ::TkEsweepXYPlot::plot IMD
}

proc showIMD {ID {noplot {}}} {
  if {$ID >= 0} {
    set name [getProp $ID Name]
    set fr [getProp $ID FR]
    set imd [getProp $ID IMD]
    set color [getProp $ID Color]

    # draw the frequency response
    if {[::TkEsweepXYPlot::exists IMD $ID]} {
      ::TkEsweepXYPlot::configTrace IMD $ID -trace $fr -color $color -width 2 -state {on} -name $name -scale y1
      ::TkEsweepXYPlot::configTrace IMD ${ID}_IMD -trace $imd -color $color -width 2 -state {on} -name ${name}_IMD -scale y1
    } else {
      ::TkEsweepXYPlot::addTrace IMD $ID $name $fr
      ::TkEsweepXYPlot::addTrace IMD ${ID}_IMD ${name}_IMD $imd
      ::TkEsweepXYPlot::configTrace IMD $ID -color $color -width 2 -scale y1
      ::TkEsweepXYPlot::configTrace IMD ${ID}_IMD -color $color -width 2 -scale y1 -dash .
    }
    if {$noplot eq {}} {
      ::TkEsweepXYPlot::plot IMD
    }
  }
}

# ---------- app windows
proc createLevelMeter {w} {
  # the level meter is a frame within a frame
  # while the outer frame keeps constant in size
  # the inner frame can be adjusted in height by calling
  # setLevelMeter
  frame $w.left  -relief raised -background red4 -bd 2
  frame $w.right  -relief raised -background red4 -bd 2

  ttk::label $w.l -text {L}
  ttk::label $w.r -text {R}

  place $w.l -relx 0 -rely 0.0 -relwidth 0.5 -anchor nw
  place $w.r -relx 0.5 -rely 0.0 -relwidth 0.5 -anchor nw
  place $w.left -relx 0 -rely 1.0 -relwidth 0.5 -relheight 0.0 -anchor sw
  place $w.right -relx 0.5 -rely 1.0 -relwidth 0.5 -relheight 0.0 -anchor sw

  # create traces on the maxR and maxL variables
  ::audio::strace add [::audio::map Left],Level,RMS write [list setLevelMeter $w.left]
  ::audio::strace add [::audio::map Right],Level,RMS write [list setLevelMeter $w.right]
}

proc setLevelMeter {w name1 name2 op} {
  upvar $name1 var
  # Color code:
  # $level < -20 dB: red (much too quiet)
  # -20 dB < $level < -6 dB: yellow (too quiet)
  # -6 dB < $level < -1 dB: green (OK)
  # -1 dB < $level: red (overload)
  set level [expr {20*log10($var($name2) < 1e-5 ? 1e-5 : $var($name2))}]
  if {$level < -64} {
    set color red4
  } elseif {$level < -48} {
    set color red
  } elseif {$level < -36} {
    set color yellow
  } elseif {$level < -1} {
    set color green
  } else {
    set color red
  }
  # configure level meter
  $w configure -background $color
  place configure $w -relheight [expr {($level+100)/100.0}]
}

proc appWindows {} {
  ####################
  # The command line #
  ####################
  ::cmdbar::init .cmdBar
  set ::cmdbar::options(Debug) 0

  set t [frame .top]
  ###########################
  # The configuration frame #
  ###########################
  set i [frame $t.input]

  ###################
  # The level meter #
  ###################
  frame $i.level -width 30 -bd 2 -relief sunken
  createLevelMeter $i.level
  pack $i.level -side right -fill y -expand no

  ##################################
  # The frequency response diagram #
  ##################################
  set f [frame $t.fr]
  set o [frame $t.o]
  pack $f -expand yes -fill both -side left
  pack $o -expand no -fill both -side left
  pack $i -fill y -side left
  pack $t -side top -fill both -expand yes
  pack .cmdBar -side top -fill x

  # create the plot
  createPlots $f
  update

  initOverlays $o

  ##############################
  # application title and icon #
  ##############################
  updateAppTitle IMD

  # constrain window size and aspect ratio
  #wm aspect . 4 3 16 9
  wm minsize . 800 600
  wm protocol . WM_DELETE_WINDOW [list :q]

  focus $t
}

proc updateAppTitle {win} {
  switch -- $win {
    IMD {
      set title {IMD }
      append title { | FS: } [expr {[setup cget Audio,Samplerate]/1000}] { kHz}  
      wm title . $title
    }
  }
}

